call_args = (
{% for arg in fortran_module_callable_args %}
    {{ arg }},
{% endfor %}
)
{{ result_name }}{{ length_suffix }} = {{ fortran_module_callable }}{{ length_callable_suffix }}(
    *call_args
)
{# TODO: re-name n in call here to {{ result_name }}{{ length_suffix }} #}
{{ result_name }}_raw = {{ fortran_module_callable }}(
    *call_args,
    n={{ result_name }}{{ length_suffix }},
)

{{ result_name }}: {{ return_type }} = {{ result_name }}_raw.decode()

return {{ result_name }}
