{{ attribute.definition.description }}

Returns
-------
    Attribute value, retrieved from Fortran.

    The value is a copy of the derived type's data.
    Changes to this value will not be reflected
    in the underlying instance of the derived type.
    To make changes to the underlying instance, use the setter instead.
